@@ -181,7 +181,8 @@ struct _global_handles {
     char vmname[32];    /* name of VM */
     int domid;        /* Xen domain id (GUI) */
     int target_domid;        /* Xen domain id (VM) - can differ from domid when GUI is stubdom */
-    uint32_t agent_version;  /* gui-agent (protocol) version */
+    uint32_t protocol_version;  /* Negotiated protocol version.  Must be uint32_t
+                                   as it is used as a protocol message. */
     char *cmdline_color;    /* color of frame */
     uint32_t label_color_rgb; /* color of the frame in RGB */
     char *cmdline_icon;    /* icon hint for WM */
@@ -232,6 +233,7 @@ struct _global_handles {
     Atom qubes_label, qubes_label_color, qubes_vmname, qubes_vmwindowid, net_wm_icon;
     bool in_dom0; /* true if we are in dom0, otherwise false */
     Atom net_supported;
+    bool permit_subwindows : 1; /* Permit subwindows */
 };
 
 typedef struct _global_handles Ghandles;
