@@ -38,7 +38,6 @@
  *
  */
 
-#define _GNU_SOURCE
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
@@ -48,6 +47,7 @@
 #include <unistd.h>
 #include <string.h>
 #include <errno.h>
+#include <stdbool.h>
 #include <fcntl.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -57,6 +57,7 @@
 #include <sys/file.h>
 #include <sys/socket.h>
 #include <sys/un.h>
+#include <err.h>
 
 #include <pulse/pulseaudio.h>
 #include <pulse/error.h>
@@ -194,27 +195,34 @@ static void start_drain(struct userdata *u, pa_stream *s) {
 
 static void process_playback_data(struct userdata *u, pa_stream *s, size_t max_length)
 {
-    int l = 0;
-    size_t index, buffer_length;
+    int l = 0, space_in_vchan;
+    size_t buffer_length;
     void *buffer = NULL;
     assert(s);
 
-    if (!libvchan_data_ready(u->play_ctrl) || !max_length)
+    if (!max_length)
         return;
 
-    buffer_length = max_length;
+    if ((space_in_vchan = libvchan_data_ready(u->play_ctrl)) < 0) {
+        pacat_log("libvchan_data_ready() failed: return value %d", space_in_vchan);
+        quit(u, 1);
+        return;
+    }
 
-    if (pa_stream_begin_write(s, &buffer, &buffer_length) < 0) {
+    buffer_length = (size_t)space_in_vchan > max_length ? max_length : (size_t)space_in_vchan;
+    if (!buffer_length)
+        return;
+
+    if (pa_stream_begin_write(s, &buffer, &buffer_length) || !buffer) {
         pacat_log("pa_stream_begin_write() failed: %s", pa_strerror(pa_context_errno(u->context)));
         quit(u, 1);
         return;
     }
-    index = 0;
 
-    while (libvchan_data_ready(u->play_ctrl) && buffer_length > 0) {
-        l = libvchan_read(u->play_ctrl, buffer + index, buffer_length);
+    if (buffer_length) {
+        l = libvchan_read(u->play_ctrl, buffer, buffer_length);
         if (l < 0) {
-            pacat_log("vchan read failed");
+            pacat_log("libvchan_read() failed: return value %d", l);
             quit(u, 1);
             return;
         }
@@ -223,12 +231,12 @@ static void process_playback_data(struct userdata *u, pa_stream *s, size_t max_l
             quit(u, 0);
             return;
         }
-        buffer_length -= l;
-        index += l;
-    }
-
-    if (index) {
-        if (pa_stream_write(s, buffer, index, NULL, 0, PA_SEEK_RELATIVE) < 0) {
+        if ((size_t)l != buffer_length) {
+            pacat_log("unexpected short vchan read: expected %zu, got %d", buffer_length, l);
+            quit(u, 1);
+            return;
+        }
+        if (pa_stream_write(s, buffer, buffer_length, NULL, 0, PA_SEEK_RELATIVE) < 0) {
             pacat_log("pa_stream_write() failed: %s", pa_strerror(pa_context_errno(u->context)));
             quit(u, 1);
             return;
@@ -249,15 +257,17 @@ static void stream_write_callback(pa_stream *s, size_t length, void *userdata) {
     process_playback_data(u, s, length);
 }
 
-static void send_rec_data(pa_stream *s, struct userdata *u) {
-    int l;
+static void send_rec_data(pa_stream *s, struct userdata *u, bool discard_overrun) {
     const void *rec_buffer;
-    size_t rec_buffer_length;
-    int rec_buffer_index;
+    size_t rec_buffer_length, rec_buffer_index = 0;
+    int l, vchan_buffer_space;
 
     assert(s);
     assert(u);
 
+    if (!u->rec_allowed)
+        return;
+
     if (pa_stream_readable_size(s) <= 0)
         return;
 
@@ -266,12 +276,30 @@ static void send_rec_data(pa_stream *s, struct userdata *u) {
         quit(u, 1);
         return;
     }
-    rec_buffer_index = 0;
+    if (!rec_buffer)
+        return;
 
-    while (rec_buffer_length > 0 && u->rec_allowed) {
-        /* can block */
+    if (u->never_block) {
+        if ((vchan_buffer_space = libvchan_buffer_space(u->rec_ctrl)) < 0) {
+            pacat_log("libvchan_buffer_space failed");
+            quit(u, 1);
+            return;
+        }
+
+        if (rec_buffer_length > (size_t)vchan_buffer_space) {
+            if (!discard_overrun)
+                return;
+            size_t bytes_to_discard = rec_buffer_length - (size_t)vchan_buffer_space;
+            pacat_log("Overrun: discarding %zu bytes", bytes_to_discard);
+            rec_buffer += bytes_to_discard;
+            rec_buffer_length = (size_t)vchan_buffer_space;
+        }
+    }
+
+    while (rec_buffer_length > 0) {
+        /* can block if u->never_block is not set */
         if ((l=libvchan_write(u->rec_ctrl, rec_buffer + rec_buffer_index, rec_buffer_length)) < 0) {
-            pacat_log("libvchan_write failed");
+            pacat_log("libvchan_write failed: return value %d", l);
             quit(u, 1);
             return;
         }
@@ -288,7 +316,7 @@ static void stream_read_callback(pa_stream *s, size_t length, void *userdata) {
     assert(s);
     assert(length > 0);
 
-    send_rec_data(s, u);
+    send_rec_data(s, u, true);
 }
 
 /* vchan event */
@@ -365,9 +393,7 @@ static void vchan_rec_callback(pa_mainloop_api *UNUSED(a),
         }
     }
     /* send the data if space is available */
-    if (libvchan_buffer_space(u->rec_ctrl)) {
-        send_rec_data(u->rec_stream, u);
-    }
+    send_rec_data(u->rec_stream, u, false);
 }
 
 /* This routine is called whenever the stream state changes */
@@ -704,16 +730,20 @@ static void control_socket_callback(pa_mainloop_api *UNUSED(a),
         ret = read(client_fd, command_buffer+command_len, sizeof(command_buffer)-command_len);
         if (ret < 0) {
             pacat_log("Control client read failed: %s", strerror(errno));
+            return;
         }
         command_len += ret;
         if (ret == 0)
             break;
-    } while (!memchr(command_buffer, '\n', command_len));
+    } while (!memchr(command_buffer + (command_len-ret), '\n', ret));
 
     if (strncmp(command_buffer, "audio-input 0\n", command_len) == 0) {
         new_rec_allowed = 0;
     } else if (strncmp(command_buffer, "audio-input 1\n", command_len) == 0) {
         new_rec_allowed = 1;
+    } else {
+        pacat_log("Invalid command buffer");
+        return;
     }
     if (new_rec_allowed != -1) {
         g_mutex_lock(&u->prop_mutex);
@@ -722,7 +752,7 @@ static void control_socket_callback(pa_mainloop_api *UNUSED(a),
         if (u->rec_allowed && u->rec_requested) {
             pacat_log("Recording start");
             pa_stream_cork(u->rec_stream, 0, NULL, NULL);
-        } else if (!u->rec_allowed &&
+        } else if (!u->rec_allowed && u->rec_stream &&
                 (u->rec_requested || !pa_stream_is_corked(u->rec_stream))) {
             pacat_log("Recording stop");
             pa_stream_cork(u->rec_stream, 1, NULL, NULL);
@@ -747,9 +777,9 @@ static int setup_control(struct userdata *u) {
         goto fail;
     }
 
-    if (snprintf(addr.sun_path, sizeof(addr.sun_path),
+    if ((size_t)snprintf(addr.sun_path, sizeof(addr.sun_path),
                 "/var/run/qubes/audio-control.%s", u->name)
-            >= (int)sizeof(addr.sun_path)) {
+            >= sizeof(addr.sun_path)) {
         pacat_log("VM name too long");
         goto fail;
     }
@@ -828,6 +858,13 @@ static void control_cleanup(struct userdata *u) {
         qdb_close(u->qdb);
 }
 
+static _Noreturn void usage(char *arg0) {
+    fprintf(stderr, "usage: %s [-l] [--] domid domname\n",
+            arg0 ? arg0 : "pacat-simple-vchan");
+    fprintf(stderr, "  -l - low-latency mode (higher CPU usage)\n");
+    exit(1);
+}
+
 int main(int argc, char *argv[])
 {
     struct timeval tv;
@@ -835,43 +872,34 @@ int main(int argc, char *argv[])
     pa_glib_mainloop* m = NULL;
     pa_time_event *time_event = NULL;
     char *server = NULL;
-    char *domname = NULL;
     int domid = -1;
     char *pidfile_path;
     int pidfile_fd;
-    int i;
+    int i = 1;
 
-    if (argc <= 2) {
-        fprintf(stderr, "usage: %s [-l] domid domname\n", argv[0]);
-        fprintf(stderr, "  -l - low-latency mode (higher CPU usage)\n");
-        exit(1);
+    if (argc <= 2 || argc > 5)
+        usage(argv[0]);
+    if (!strcmp(argv[i], "-l")) {
+        bufattr = &custom_bufattr;
+        i++;
     }
-    for (i=1; i<argc; i++) {
-        if (argv[i][0] == '-') {
-            switch (argv[i][1]) {
-                case 'l':
-                    bufattr = &custom_bufattr;
-                    break;
-                default:
-                    fprintf(stderr, "Invalid option: %c\n", argv[i][1]);
-                    exit(1);
-            }
-        } else {
-            if (domid < 0)
-                domid = atoi(argv[i]);
-            else 
-                domname = argv[i];
-        }
-    }
-    if (domid < 0) { /* not-a-number returns 0 */
-        fprintf(stderr, "invalid domid\n");
-        exit(1);
-    }
-    if (!domname) {
-        fprintf(stderr, "missing domname\n");
-        exit(1);
-    }
-
+    if (!strcmp(argv[i], "--"))
+        i++;
+    if (argc - i != 2)
+        usage(argv[0]);
+    const char *domid_str = argv[i], *domname = argv[i + 1];
+    char *endptr;
+    errno = 0;
+    long l_domid = strtol(domid_str, &endptr, 10);
+    /* 0x7FF0 is DOMID_FIRST_RESERVED */
+    if (l_domid < 0 || l_domid >= 0x7FF0 || errno == ERANGE)
+        errx(1, "domid %ld out of range 0 through %d inclusive", l_domid,
+                0x7FF0 - 1);
+    domid = l_domid;
+    if (errno)
+        err(1, "invalid domid %s", domid_str);
+    if (*endptr)
+        errx(1, "trailing junk after domid %s", domid_str);
     if (create_pidfile(domid, &pidfile_path, &pidfile_fd) < 0)
         /* error already printed by create_pidfile() */
         exit(1);
@@ -901,7 +929,7 @@ int main(int argc, char *argv[])
         perror("setuid");
         exit(1);
     }
-
+    u.never_block = libvchan_buffer_space(u.rec_ctrl) > 8192;
     u.proplist = pa_proplist_new();
     pa_proplist_sets(u.proplist, PA_PROP_APPLICATION_NAME, u.name);
     pa_proplist_sets(u.proplist, PA_PROP_MEDIA_NAME, u.name);
